/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                              boOX 0-047_dukla                              */
/*                                                                            */
/*                      (C) Copyright 2018 Pavel Surynek                      */
/*                http://www.surynek.com | <pavel@surynek.com>                */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
/* cbs.h / 0-047_dukla_047                                                    */
/*----------------------------------------------------------------------------*/
//
// Conflict based search implemented in a standard way. A version for MAPF and
// a version for the token swapping have been implemented.
//
/*----------------------------------------------------------------------------*/


#ifndef __SMTCBS_H__
#define __SMTCBS_H__

#include <vector>
#include <list>
#include <set>
#include <map>
#include <unordered_map>
#include <unordered_set>

#include "types.h"
#include "result.h"

#include "core/graph.h"
#include "core/agent.h"


using namespace std;


/*----------------------------------------------------------------------------*/

namespace boOX
{


/*----------------------------------------------------------------------------*/
// sCBS

    class sCBS
    {
    public:
	typedef std::vector<sInt_32> VertexIDs_vector;
	typedef std::list<sInt_32> VertexIDs_list;
	typedef std::unordered_set<sInt_32> VertexIDs_uset;
	
	typedef std::vector<VertexIDs_uset> Conflicts_vector;
	
	typedef std::vector<Conflicts_vector> AgentConflicts_vector;
	typedef std::vector<VertexIDs_vector> AgentPaths_vector;

    public:
	struct Visit
	{
	    Visit() { /* nothing */ }
	    Visit(sInt_32 level, sInt_32 vertex_id, sInt_32 previous_id)
	    : m_level(level), m_vertex_id(vertex_id), m_previous_id(previous_id) { /* nothing */ }
	    
	    sInt_32 m_level;
	    sInt_32 m_vertex_id;
	    sInt_32 m_previous_id;
	};

	struct Conflict
	{
	    Conflict() { /* nothing */ }
	    Conflict(sInt_32 vertex_id, sInt_32 agent_id, sInt_32 level)
	    : m_vertex_id(vertex_id), m_agent_id(agent_id), m_level(level) { /* nothing */ }

	    sInt_32 m_vertex_id;
	    sInt_32 m_agent_id;
	    sInt_32 m_level;
	};

	struct Node
	{
	    Node(sInt_32 upd_agent_id)
	    : m_upd_agent_id(upd_agent_id) { /* nothing */ }
	    Node(sInt_32 upd_agent_id, const AgentConflicts_vector agent_Conflicts, const AgentPaths_vector agent_Paths)
	    : m_upd_agent_id(upd_agent_id), m_agent_Paths(agent_Paths), m_agent_Conflicts(agent_Conflicts) { /* nothing */ }

	    sInt_32 m_upd_agent_id;
	    
	    AgentPaths_vector m_agent_Paths;
	    AgentConflicts_vector m_agent_Conflicts;
	};

	typedef std::unordered_map<sInt_32, Visit> Visits_umap;
	typedef std::vector<Visits_umap> Visits_vector;
	typedef std::list<Visit> Visits_list;

	typedef std::unordered_map<sInt_32, sInt_32> Occupation_umap;
	typedef std::vector<Occupation_umap> Occupations_vector;

	typedef std::map<sInt_32, Node> Nodes_map;
	typedef std::multimap<sInt_32, Node, std::less<sInt_32>> Nodes_mmap;	
	
    public:
	sCBS(sInstance *instance);
	/*----------------------------------------------------------------------------*/

	sInt_32 find_ShortestNonconflictingSwapping(sSolution &solution, sInt_32 cost_limit) const;
	sInt_32 find_ShortestNonconflictingSwapping(const sInstance &instance, sSolution &solution, sInt_32 cost_limit) const;

	sInt_32 find_ShortestNonconflictingSwapping(AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const;	
	sInt_32 find_ShortestNonconflictingSwapping(const sInstance   &instance,
						    AgentPaths_vector &agent_Paths,
						    sInt_32            cost_limit) const;	
	
	sInt_32 find_ShortestNonconflictingPaths(AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const;	
	sInt_32 find_ShortestNonconflictingPaths(const sInstance   &instance,
						 AgentPaths_vector &agent_Paths,
						 sInt_32            cost_limit) const;

	sInt_32 find_NonconflictingSwapping(AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const;	
	sInt_32 find_NonconflictingSwapping(const sInstance   &instance,
					    AgentPaths_vector &agent_Paths,
					    sInt_32            cost_limit) const;	

	sInt_32 find_NonconflictingPaths(AgentPaths_vector &agent_Paths, sInt_32 cost_limit) const;	
	sInt_32 find_NonconflictingPaths(const sInstance   &instance,
					 AgentPaths_vector &agent_Paths,
					 sInt_32            cost_limit) const;

	sInt_32 find_NonconflictingSwapping_base(const sInstance       &instance,
						AgentConflicts_vector &agent_Conflicts,
						AgentPaths_vector     &agent_Paths,
						sInt_32                cost_limit) const;

	sInt_32 find_NonconflictingSwapping_upd(const sInstance       &instance,
						AgentConflicts_vector &agent_Conflicts,
						AgentPaths_vector     &agent_Paths,
						sInt_32                cost_limit) const;

	sInt_32 find_NonconflictingSwapping_prio(const sInstance       &instance,
						 AgentConflicts_vector &agent_Conflicts,
						 AgentPaths_vector     &agent_Paths,
						 sInt_32                cost_limit) const;		

	sInt_32 update_NonconflictingSwapping(sInt_32                upd_agent_id,
					      const sInstance       &instance,
					      Occupations_vector    &space_Occupations,
					      AgentConflicts_vector &agent_Conflicts,
					      AgentPaths_vector     &agent_Paths,
					      sInt_32                cost_limit) const;

	sInt_32 revise_NonconflictingSwapping(const sInstance         &instance,
					      AgentConflicts_vector   &agent_Conflicts,
					      const AgentPaths_vector &agent_Paths,
					      sInt_32                  cost_limit,
					      Nodes_mmap              &search_Queue) const;
	/*----------------------------------------------------------------------------*/	
		
	sInt_32 find_NonconflictingPaths(const sInstance       &instance,
					 AgentConflicts_vector &agent_Conflicts,
					 AgentPaths_vector     &agent_Paths,
					 sInt_32                cost_limit) const;

	sInt_32 find_NonconflictingSequence(const sUndirectedGraph &graph,
					    sInt_32                 source_id,
					    sInt_32                 sink_id,
					    const Conflicts_vector &Conflicts,
					    VertexIDs_vector       &Path) const;

	void cast_Occupations(sInt_32 agent_id, sInt_32 prefix_length, const VertexIDs_vector &path, Occupations_vector &space_Occupations) const;
	void uncast_Occupations(sInt_32 prefix_length, const VertexIDs_vector &path, Occupations_vector &space_Occupations) const;

    private:
	sCBS(const sCBS &cbs);
	const sCBS& operator=(const sCBS &cbs);
	
    public:
	sInstance *m_Instance;

	Conflicts_vector m_Conflicts;
    };



/*----------------------------------------------------------------------------*/

} // namespace boOX

#endif /* __CBS_H__ */
